# -*- coding: utf-8 -*

import os, json
from ctypes import cdll
# import sys
# sys.path.append('../')

from Log.Logger import log2File

class PksManager:
    def __init__(self, install_dir):
        self.install_dir = install_dir
        self.httc_lib = None
        self.kysec_lib = None

        self.__load_httclib()
        self.__load_kyseclib()

    def __format_cpu_arch(self):
        mach_archs = {
            "x86": "x86",
            "i386":"x86",
            "i586": "x86",
            "i686": "x86",
            "x86_64":"amd64",
            "aarch64":"arm64",
            "mips64el":"mips64",
            "sw64":"sw64"
        }

        mach = os.uname()[4]
        if mach in mach_archs:
            arch = mach_archs[mach]
        else:
            arch = mach

        return arch

    def __do_get_dlpath(self,sect_items):
        dlpath = ''
        arch = self.__format_cpu_arch()
        if not 'dlpath' in sect_items:
            return dlpath

        if arch in sect_items['dlpath']:
            dlpath = sect_items['dlpath'][arch]

        return dlpath

    def get_dlpath(self,section):
        conf_file = os.path.join(self.install_dir, 'conf/pks.json')
        if not os.path.exists(conf_file):
            log2File(conf_file + ' is not exist.')
            return ''
        with open(conf_file, 'r') as f:
            data = json.load(f)
            if section in data:
               return self.__do_get_dlpath(data[section])
            else:
                log2File(conf_file+': node[' + section + ']is not exit.')
        return ''

    def __load_dylib(self,name):
        dllib = None
        dlpath = self.get_dlpath(name)
        if dlpath == '':
            return dllib
        try:
            dllib = cdll.LoadLibrary(dlpath)
        except Exception as e:
            log2File(str(e.args))

        return dllib

    def __load_httclib(self):
        dllib = self.__load_dylib('httc')
        if dllib is not None:
            self.httc_lib = dllib

    def __load_kyseclib(self):
        dllib = self.__load_dylib('kysec')
        if dllib is not None:
            self.kysec_lib = dllib

    def __read_file(self,fname):
        data = ''

        try:
            with open(fname, 'r') as f:
                data = f.read()
        except Exception as e:
            log2File('read file' + fname + 'error: ' + str(e.args))

        return data

    def __doHttcVerify(self,datawhl_file,datasig_file):
        #此种情况就不用加白了，返回true
        if self.httc_lib is None:
            return True

        #两个文件必须同时存在才行,一个不存在我们都主粉是无需要加白
        if not os.path.exists(datawhl_file):
            slog = ("httc: file %s is not exist; " +
                "we suppose don't need verify") % datawhl_file
            log2File(slog)
            return True

        if not os.path.exists(datasig_file):
            slog = ("httc: file %s is not exist; " +
                "we suppose don't need verify") % datasig_file
            log2File(slog)
            return True

        datawhl = self.__read_file(datawhl_file)
        if datawhl_file == '':
            slog = "httc: read file %s failed; " % datawhl_file
            log2File(slog)
            return False

        datasig = self.__read_file(datasig_file)
        if datasig == '':
            slog = "httc: read file %s failed; " % datasig_file
            log2File(slog)
            return False

        try:
            rc = self.httc_lib.HTTC_VerifyAndApplyWhitelist(datawhl,len(datawhl),
                                        datasig,len(datasig))
            if rc != 0 :
                log2File("HttcVerify call error,rc: %d" % rc)
            return (rc == 0)
        except Exception as e:
            log2File('HttcVerify error: ' + str(e.args))
            return False

    def HttcVerify(self,tmp_update_dir):
        #此种情况就不用加白了，返回true
        if self.httc_lib is None:
            return True

        #病毒库的与主程序的有一个就认为是成功
        engine_whl_file = os.path.join(tmp_update_dir,"engine_httc_whl")
        engine_sig_file = os.path.join(tmp_update_dir,"engine_httc_sig")

        #文件不存在时httcVerify会返回True
        ok = self.__doHttcVerify(engine_whl_file,engine_sig_file)
        if not ok: return ok

        main_whl_file = os.path.join(tmp_update_dir,"main_httc_whl")
        main_sig_file = os.path.join(tmp_update_dir,"main_httc_sig")

        ok = self.__doHttcVerify(main_whl_file,main_sig_file)
        return ok

    def KySecUpdate(self,fname):
        #此种情况就不用加白了，返回true
        if self.kysec_lib is None:
            log2File("KySec don't need update for file " + fname)
            return True

        if not os.path.exists(fname):
            log2File("KySec file %s is not exist" % fname)
            return True

        try:
            '''接口的返回值:
                0： 更新成功
                1： 认证失败，没有权限
                -1：其他失败
                -2：参数错误，路径不存在或参数不对
                -3：文件类型不对，非可执行文件、shell脚本、python脚本、so动态链接库、ko内核模块
            '''
            #只要返回值不是1我们就认为是成功,这里一定要是字符串
            rc = self.kysec_lib.kysec_whlist_exectl_update(str(fname))
            if rc == 1 :
                log2File(("KySec update whitelist for file %s failed,err: %d ") % 
                        (fname,rc))

            return (rc != 1)
        except Exception as e:
            log2File('KySecUpdate error: ' + str(e.args))
            return False

'''
if __name__ == "__main__":
    install_dir = "./"
    mgr = PksManager(install_dir)
    dlpath = mgr.get_dlpath('httc')
    print dlpath

    dlpath = mgr.get_dlpath('kysec')
    print dlpath

    mgr.HttcVerify('./')
    mgr.KySecUpdate('main_httc_sig')
'''
